#include "../bayos.h"

Task** Scheduler::task_list  = NULL;
List<Task*>* Scheduler::runq = NULL;
Task* Scheduler::current     = NULL;
bool Scheduler::ready        = false;

/* XPW[̏ */
void Scheduler::init() {
    /* ϐ̏ */
    Scheduler::task_list = new Task*[MAX_TASKS];
    Scheduler::runq      = new List<Task*>();

    /* ^XNꗗ̏ */
    SegmentDescriptor* gdt = (SegmentDescriptor *) ADDRESS_GDT;
    for (int i = 0; i < MAX_TASKS; i++) {
        Scheduler::task_list[i] = new Task();
        Scheduler::task_list[i]->id       = i;
        Scheduler::task_list[i]->selector = (TASK_GDT_0 + i) * 8;
        Scheduler::task_list[i]->enable   = false;
        Scheduler::task_list[i]->fifo     = new List<Message*>();
        Scheduler::task_list[i]->tss.ldtr = (TASK_GDT_0 + MAX_TASKS + i) * 8;
        /* J[l^XNpfXNv^̐ݒ */
        Descriptor::setSegmentDescriptor
        (gdt + TASK_GDT_0 + i, 103, (int) &Scheduler::task_list[i]->tss, AR_TSS32);
        /* AvP[VpfXNv^̐ݒ */
        Descriptor::setSegmentDescriptor
        (gdt + TASK_GDT_0 + MAX_TASKS + i, 15, (int) Scheduler::task_list[i]->ldt, AR_LDT);
    }

    /* [0]J[l^XN̐ݒ */
    Scheduler::current = allocateTask(0, "KERNEL");

    /* [1]ACh^XN̐ݒ */
    Task* idle = allocateTask((int) &task_idle, "IDLE");

    /* [2]L[{[h^XN̐ݒ */
    Task* keyboard = allocateTask((int) &task_keyboard, "KEYBOARD");

    /* [3]}EX^XN̐ݒ */
    Task* mouse = allocateTask((int) &task_mouse, "MOUSE");

    /* [4]R\[^XN̐ݒ */
    Task* console = allocateTask((int) &task_console, "CONSOLE");

    /* sL[Ƀ^XNlߍ */
    Scheduler::runq->add(idle);
    Scheduler::runq->add(keyboard);
    Scheduler::runq->add(mouse);
    Scheduler::runq->add(console);
    Scheduler::runq->add(Scheduler::current);

    /* XPW[O\ */
    Scheduler::ready = true;
}

/* 󂢂Ă^XN̎擾 */
Task* Scheduler::allocateTask(int eip, const char* name) {
    for (int i = 0; i < MAX_TASKS; i++) {
        if (Scheduler::task_list[i]->enable == false) {
            Scheduler::task_list[i]->enable = true;
            memcpy(task_list[i]->name, name, strlen(name) + 1);
            if (i == 0) {
                load_tr(Scheduler::task_list[i]->selector);
            } else {
                Scheduler::task_list[i]->tss.esp     = (int) (new byte[64 * 1024] + (64 * 1024));
                Scheduler::task_list[i]->tss.eip     = eip;
                Scheduler::task_list[i]->tss.eflags  = 0x00000202; /* IF = 1; */
                Scheduler::task_list[i]->tss.eax     = 0; /* Ƃ肠0ɂĂƂɂ */
                Scheduler::task_list[i]->tss.ecx     = 0;
                Scheduler::task_list[i]->tss.edx     = 0;
                Scheduler::task_list[i]->tss.ebx     = 0;
                Scheduler::task_list[i]->tss.ebp     = 0;
                Scheduler::task_list[i]->tss.esi     = 0;
                Scheduler::task_list[i]->tss.edi     = 0;
                Scheduler::task_list[i]->tss.es      = 1 * 8;
                Scheduler::task_list[i]->tss.cs      = 2 * 8;
                Scheduler::task_list[i]->tss.ss      = 1 * 8;
                Scheduler::task_list[i]->tss.ds      = 1 * 8;
                Scheduler::task_list[i]->tss.fs      = 1 * 8;
                Scheduler::task_list[i]->tss.gs      = 1 * 8;
                Scheduler::task_list[i]->tss.iomap   = 0x40000000;
                Scheduler::task_list[i]->tss.ss0     = 0;
            }
            return Scheduler::task_list[i];
        }
    }
    return NULL;
}

/* ^XNE */
void Scheduler::killTask(int id) {
    /* J[l^XNR\[^XN܂ł͎EȂ */
    if (TASK_CONSOLE < id && id < MAX_TASKS) {
        Task* task = Scheduler::task_list[id];
        task->enable = false;
        task->fifo->removeAll();
        Scheduler::runq->remove(task);
        switchTask();
    }
}

/* Iȃ^XNXCb` */
void Scheduler::switchTask() {
    if (Scheduler::ready == false) return;
    Scheduler::current = Scheduler::runq->removeAt(0);
    Scheduler::runq->add(Scheduler::current);
    farjmp(0, Scheduler::current->selector);
}

/* ACh^XN */
void task_idle(void) {
    for (;;) {
        io_hlt();
    }
}
